package Narada;

use strict;
use warnings;

use base 'Exporter';

use CGI::Carp qw(carp confess);
use DBI;
use Config::YAML;
use Cache::Memcached::libmemcached;
use Digest::MD5 qw(md5_hex);
use Data::Dumper;
use Apache2::Directive;
use Apache2::Cookie;
use Apache2::Request;
use Apache2::Upload ();
use Storable qw(nfreeze thaw);
use Mail::Sendmail;
use Email::Valid;
use HTML::Entities;
use File::Slurp;

our $VERSION = '0.200';
our @EXPORT = qw(randomString buildRange sendEmail $DEBUG $COOKIEDOMAIN 
$COOKIENAME $COOKIEPATH $MEMC_SERVERS $GMAN_SERVERS $MEMC_CATALOGUE_EXPIRERY $DB_USER
$DB_PASSWORD $MEMC_CATALOGUE_EXPIREY $DB_DRIVER $DB_SCHEMA 
$GEARMAN_SERVERS $URL_STORE_COUNT $SPHINX_REINDEX_COUNT);

# defaults

# Make sure this is set correctly!
our $NARADA_HOME = $ENV{NARADA_HOME} || '/home/narada/narada';

our $DB_DRIVER = 'mysql';
our $DB_SCHEMA = 'narada';
our $DEBUG = 0 ;
our $COOKIENAME = 'websession';
our $COOKIEDOMAIN = `hostname`; 
our $COOKIEPATH= '/';
our $URL_STORE_COUNT = 20;
our $SPHINX_REINDEX_COUNT = 30;
our $DB_USER = 'narada';
our $DB_PASSWORD = '';

# memcached server list, can be overridden upon instantiation
our $MEMC_SERVERS= ['localhost:11211'];

# gearman server list
our $GMAN_SERVERS = ['127.0.0.1:4730'];

# when to expire the cache catalogue
our $MEMC_CATALOGUE_EXPIREY= 300;

# all columns except primary key for the user, which auto_increment sets
our $USER_COLS = [qw(
        username 
        password
        email
        firstname
        surname
        phone
        city
        state
        address
        )]; 

# required columns for submissions
our $REQUIRED_COLS = [ qw(username email firstname surname password)]; 


=head2 new() 

instantiates a new WebApp object

=item args

=over

$caller - class name

$opts - for overriding default values

=back
 
=item returns

=over

$self - object reference

=back

=cut

sub new {
  my($class, $opts) = @_;

  my $config = Config::YAML->new(config => "/home/narada/narada/conf/narada.conf");

  # memcached array ref of servers
  @$MEMC_SERVERS = $config->{memc_servers} ?
    split(',', $config->{memc_servers}) : $MEMC_SERVERS;

  # gearman array ref of servers
  @$GMAN_SERVERS = $config->{gman_servers} ?
    split(',', $config->{gman_servers}) : $GMAN_SERVERS;

  $DB_USER = $config->{db_user} ?
    $config->{db_user} : $DB_USER;
  $DB_PASSWORD = $config->{db_password} ?
    $config->{db_password} : $DB_PASSWORD;
    
  $MEMC_CATALOGUE_EXPIREY = $config->{memc_catalogue_expirey} ?
    $config->{memc_catalogue_expirey} : $MEMC_CATALOGUE_EXPIREY;

  $COOKIENAME = $config->{cookiename} ?
    $config->{cookiename} : $COOKIENAME;

  $COOKIEPATH = $config->{cookiepath} ?
    $config->{cookiepath} : $COOKIEPATH;

  $COOKIEDOMAIN = $config->{cookiedomain} ?
    $config->{cookiedomain} : $COOKIEDOMAIN;

  $URL_STORE_COUNT = $config->{url_store_count} ?
    $config->{url_store_count} : $URL_STORE_COUNT;

  $SPHINX_REINDEX_COUNT = $config->{sphinx_reindex_count} ?
    $config->{sphinx_reindex_count} : $SPHINX_REINDEX_COUNT;

  $DEBUG = $config->{debug} ? $config->{debug} : 0;

  my $self = {
    db_user             => $DB_USER,
    db_password         => $DB_PASSWORD,
    memc_servers        => $MEMC_SERVERS,
    gman_servers        => $GMAN_SERVERS,

    # set the message to empty string
    msg => '',
  };
  $self->{$_} = $opts->{$_} for keys %$opts;

  bless($self, $class);

  if ($self->{r}) {
      $self->{req}||= new Apache2::Request($self->{r});
      $self->{form}->{$_} = $self->{req}->param($_) for $self->{req}->param();
  }

  # connect to MySQL
  return $self unless $self->_connectDB();

  # connect to Memcached
  $self->_connectMemcached();

  return $self;
}


=head2 userExistsInCache()
 
=item args 

=over

$self - object reference

$user - user hashref

=back

=item returns

=over

uid value, if exists

0 if not 

=back

=cut

sub userExistsInCache {
  my ($self, $user)= @_;
  return 0 unless defined $self->{memc};
  # over-assuming I am
  $self->{msg}= "ERROR: The ";

  #
  # try to see if in memcached, using username or email
  # which are two possible storage keys
  #
  for my $key(qw (username email)) {
      exists $user->{$key} or next;
      my $uid= $self->{memc}->get($key);
      $uid or next;
      # provides a message showing column and value
      $self->{msg} .= "\u$key $user->{$key} exists already.";
      return $uid;
  }
  return 0;
}

=head2 userExists()

checks to see if a user exists by either the username or email address 

=item arguments

=over

$user - hashref representing a user keyed by user fields 

=back

=item returns

=over

uid > 1 - user exists 
0 - user doesn't exist 
-1 - error

=back

=cut

sub userExists {
  my ($self, $user)= @_;
  my @bind_values;

  # try memcached first before hitting the database
  my $uid= $self->userExistsInCache($user);
  return $uid if $uid;

  # ok, not in memcached, check MySQL

  # now attempt to see if in MySQL, using username or email
  for my $key(qw (username email)) {
    #
    # have to run a separate query for each column
    # otherwise there is no way to tell the user which column
    # there is a duplicate on
    #
    if (exists $user->{$key}) {
      my $query= "SELECT uid FROM users WHERE $key = ?";
      # prepare the query
      my $exists= $self->{dbh}->selectcol_arrayref($query, {}, $user->{$key});
      if ($exists->[0]) {
        $self->{msg}= "Error the \u$key $user->{$key} exists already"; 
        return ($exists->[0]);
      }
    }
  }
  return 0;
}

=head2 getUser()

obtains a user hashref, using getUsers to do the heavy lifting
taking the list of hashrefs returned (which there should only be one)
and returning a single user hashref obtained by user key uid

=item arguments: 

=over

$self - object reference

$uref - hashref with uid, username , email or other lookup value 
 
=back

=head returns:

=over

hashref representing a user keyed by user fields 

undef - no user or an error, msg will be set 

=back

=cut

sub getUser  {
  my ($self, $uref)= @_;

  # get the big user ref, keyed by uid
  $uref->{uid}= $self->userExists($uref) unless $uref->{uid};
  my $user= $self->getUsers($uref);
  # return a single hashref obtained using uid
  if ($uref->{uid}) {
    return $user->{$uref->{uid}};
  }
  return {};

}



=head2 saveUser() 

either inserts or updates the user specified by $user into both
MySQL and memcached

=item arguments

=over

$user - hashref representing a user keyed by user fields 

$msg  - scalar ref for writing the error or success message to

=back
 
=item returns

=over

1 - success 

0 - error

=back

=cut

sub saveUser {
  my ($self, $user) = @_; 

  #
  # if $user has a uid update
  #
  if ($user->{uid}) {
    my $update_user;

    # check if the user is in the database, even though a uid is specified
    # this $check_user ref will be used to see what columns actually were
    # changed 
    my $check_user= $self->getUser({ uid => $user->{uid}});

    # if the user's email address changed, make sure the address they are changing
    # it to is not already existing.
    if (defined $check_user && $check_user->{email} ne $user->{email}) {
        my $exists_uid= $self->userExists({email => $user->{email}});
        return 0 if $exists_uid == -1; 
        # let the user know they have to use a different email address
        if ($exists_uid && $exists_uid != $user->{uid}) {
          $self->{msg}=
            "ERROR: You cannot use $user->{email} as an email address, it's being currently used";
          return 0;
        }
    }
    # for this application, do not allow the user to change their username
    if ($check_user->{username} ne $user->{username}) {
      $self->{msg}= "ERROR: You cannot change your username!\n";
      return 0;
    }

    # use the user columns as the mapping
    for my $key (@$USER_COLS) {
      # tmp user ref for storing changed keys
      # only update keys that are changed
      if ($user->{$key} ne $check_user->{$key}) {
        $update_user->{$key}= $user->{$key} 
      }
    }
    # the uid has to be in $update_user in order by the WHERE clause to work
    $update_user->{uid}= $user->{uid};

    # if $update_user has keys, something changed with the user, so update
    if (keys %$update_user) {
      if ($self->updateUser($update_user)) {
        $self->{msg}= "Updated $user->{username}\n";
        # save user in cache
        $self->saveUserToCache($user);
        return 1;
      }
    }
  }
  # otherwise, no uid, must be a new user
  else {
    # make sure user doesn't exist and no error
    unless ($self->userExists($user)) {
      # insert the suer
      $user->{uid}= $self->insertUser($user);
      # if there is a $user->{uid}, then the insert was successful, so 
      # store in memcached too
      if ($user->{uid}) {
        $self->saveUserToCache($user);
        $self->{msg}= "Created $user->{username}\n";
        return 1;
      }
    }
  }
  # default return 0, if errors, msg contains error message
  return 0;
}



=head2 insertUser()

insert the user into MySQL

=item args

=over

$self - object reference

$user - user hash reference

=back

=item returns

=over

insert id containing new user uid

0 failure

=back

=cut

sub insertUser  {
  my ($self, $user)= @_;

  # build up the columns specification
  my $query = 'INSERT INTO users (' . join(',', @$USER_COLS) . ')';
  # build up the placeholders
  $query .= ' VALUES ( ' . join( ',', ( '?' ) x @$USER_COLS ) . ')'; 

  my $sth= $self->{dbh}->prepare($query);
  # use hash slice to provide the bind values, 
  # which will be in the correct order
  $sth->execute(@{$user}{@$USER_COLS});
  if ($self->{dbh}->err) {
    $self->{msg}= "Database Error: $DBI::err_str";
    return 0;
  }

  # return the new uid
  return $self->{dbh}->{mysql_insertid};
}


=head2 updateUser()

update the user in MySQL, make sure there is a uid
 
=item args

=over

$self - object reference

$user - user reference

=back

=item returns

=over

number of rows on update - should be 1 on success

0 on failure

=back

=cut

sub updateUser {
  my ($self, $user)= @_;
  my @bind_values;

  # no need to update if all there is is a uid
  if ($user->{uid} && keys %$user == 1) {
    $self->{msg}= "Contact unchanged, not updated";
    return 0;
  }

  # no uid, set error, return
  unless ($user->{uid}) {
    $self->{msg}= "ERROR: uid not set!";
    carp $self->{msg};
    return 0;
  }

  my $query = "UPDATE users SET " ;
  # construct a query
  for my $key (@$USER_COLS) {
    if (defined $user->{$key}) {
      $query .= "$key = ?,";
      push(@bind_values, $user->{$key});
    }
  }
  # chop off trailing ,
  chop($query);

  # where clause
  $query .= " WHERE uid = ?";

  # value for where clause
  push(@bind_values, $user->{uid});

  my $sth= $self->{dbh}->prepare($query);
  $sth->execute(@bind_values);
  if ($self->{dbh}->err) {
    $self->{msg}= "Database Error: $DBI::err_str";
    return 0;
  }
  # return 1 - since uid specified, only one row should be updated
  return $sth->rows();
}

=head2 deleteUserFromCache()

deletes a user from memcached - any entry in memcached for the user
including the email, username, as well as from the list of user uids

=over args

$self - object reference
$uid - user id 

=back

=over

=item returns

no value 

=back

=cut

sub deleteUserFromCache {
    my ($self, $uid)= @_;
    return 0 unless defined $self->{memc};
    my $key= 'uid_' . $uid;

    # obtain the user - why? Need the email address and username
    my $user= $self->{memc}->get($key);

    # delete entry keyed by email
    $self->{memc}->delete($user->{email});
    # delete entry keyed by username 
    $self->{memc}->delete($user->{username});

    # now delete the user
    $self->{memc}->delete($key);

    # take user out of uid list
    $self->deleteMemcUIDList($user->{uid});
}

=head2 deleteUsers()

deletes one or more users specified by uid in $delete_ids

=item arguments

=over

$self - object reference

$delete_ids - hashref keyed by uid, values usernames

=back

=item returns

=over

1 - success 

0 - no deletion or error

=back

=cut
sub deleteUsers {
  my ($self, $delete_ids) = @_; 
  # no ids, don't bother
  return 0 unless keys %$delete_ids;

  # start the query
  my $query = 'DELETE FROM users WHERE uid IN (';
  # put together placeholder list
  for my $uid(keys %$delete_ids) {
    $self->deleteUserFromCache($uid);
  }
  # build up placeholders list
  $query .= join( ',', ( '?' ) x keys %$delete_ids );
  $query .= ')';

  my $sth= $self->{dbh}->prepare($query);
  # keys are the ids
  $sth->execute(keys %$delete_ids);;
  if ($self->{dbh}->err) {
    $self->{msg}= "Database Error: $DBI::err_str";
    return 0;
  }

  # if any rows were deleted from the database
  # create a success message
  if ($sth->rows()) {
    # values are the usernames. Make a nice message
    $self->{msg}= sprintf("Deleted contacts: \u%s", join(', ', values %$delete_ids));

    return $sth->rows();
  }
  return 0;
}



=head2 getUsers()

select and retrieve an arrayref of users from the database

=item arguments

=over

$self - object reference

$uref -  user ref containing one or more keys

$reload_cache - used to cause the cached users to be reloaded

=back
 
=item returns

=over

return arrayref containing result set

=back

=cut

sub getUsers  {
  my ($self, $uref, $reload_cache)= @_; 
  my $userlist_ref;
  my @bind_values= ();

  unless ($reload_cache) {
    # obtain from cache, use if there are entries 
    $userlist_ref= $self->getUsersFromCache();
    return $userlist_ref if keys %$userlist_ref;
  }

  # not in memcached, procede to check the database

  # this check is here to make sure something happens so the page will 
  # print, as well as provide an error message
  unless (defined $self->{dbh} && $self->{dbh}->ping) {
    $self->{msg}= "Unable to connect to MySQL";
    carp $self->{msg};
    return undef;
  }
  # start a query
  my $query= 'SELECT * FROM users';

  # if the $uref hashref has any keys specified, then construct a WHERE clause 
  if (keys %$uref) {
    $query .= ' WHERE ';
    # here, we want the uid 
    for my $key('uid', @$USER_COLS) {
      if (defined $uref->{$key}) {
        $query .= "$key = ? AND ";
        push(@bind_values, $uref->{$key});
      }
    }
    # chop off trailing ' AND '
    $query= substr($query, 0, -5);
  }

  my $sth= $self->{dbh}->prepare($query);
  $sth->execute(@bind_values);
  if ($self->{dbh}->err) {
    # build a message
    $self->{msg}= "ERROR: $DBI::errstr";
    # print to log
    carp $self->{msg};
    return undef;
  }
  # force results into arrayref of hashrefs
  $userlist_ref= $sth->fetchall_hashref('uid');
  $self->cacheUsers($userlist_ref);
  return $userlist_ref;
}

 
=head2 cacheUsers()

stores the complete user reference of users to memcached

=item args

=over

$self - object reference

$uref - hashref of user hash refs

=back

=item returns

=over

success if stored

=back

=cut

sub cacheUsers {
  my ($self, $uref)= @_;
  return 0 unless defined $self->{memc};
  my $uid_list;
  for my $uid (keys %$uref) {
    # first, set the entire user, keyed by uid_<uid>
    my $key= "uid_$uid";
    $self->{memc}->set($key, $uref->{$uid});
    # if new, add it to the list, otherwise no effect
    $uid_list->{$key} = $uid ;

    # then by email 
    $self->{memc}->set($uref->{$uid}{email}, $uid);

    # then by username
    $self->{memc}->set($uref->{$uid}{username}, $uid);
    
  }
  # so we know what keys we have without asking the database!
  $self->setMemcUIDList($uid_list);
}


=head2 getMemcUIDList()

convenience method used to retrieve the UID hashref 
stored as uid_list in memcached, ensuring if it isn't 
stored either due to expiration or restart of a memcached server,
it is re-cached, then returned to the caller

=item args

=over

$self - object reference

=back

=item returns

=over

hashref 

=back

=cut
 
sub getMemcUIDList {
  my ($self)= @_;
  return 0 unless defined $self->{memc};

  # try to obtain the uid_list
  my $uid_list= $self->{memc}->get('uid_list');

  # force re-caching of users if the list is not there
  unless (ref $uid_list) {
    $self->getUsers({});
  }

  # now get the $uid_list again, should be there ;)
  $uid_list= $self->{memc}->get('uid_list');
  
  return $uid_list;
}


=head2 setMemcUIDList()

convenience method to set the uid_list object in memcached

=item args

=over

$self - object reference

$uid_list - hash reference of UIDs

=back

=item returns

=over

true - success

false - failure

=back

=cut

sub setMemcUIDList {
  my ($self, $uid_list)= @_;
  return 0 unless defined $self->{memc};
  $self->{memc}->set("uid_list", $uid_list, $MEMC_CATALOGUE_EXPIREY);

}


=head2 deleteMemcUIDList()

convenience method to updateMemcUIDList, passing the delete flag,
resulting in the UID being removed from uid_list and re-stored in
memcached

=item args

=over

$self - object reference

$uid -  UID

=back

=item returns

=over

true - success

false - failure

=back

=cut

sub deleteMemcUIDList {
  my ($self, $uid)= @_;
  $self->updateMemcUIDList($uid, 1);

}


=head2 updateMemcUIDList()

convenience method used to either insert, update or delete (if flag set)
a UID into/from the hashref stored as uid_list and re-stored in  memcached

=item args

=over

$self - object reference

uid -  UID

$delete_flag - 1 delete, 0 no delete

=back

=item returns

=over

true - success

false - failure

=back

=cut

sub updateMemcUIDList {
  my ($self, $uid, $delete_flag)= @_;
  return 0 unless defined $self->{memc};

  # if no uid supplied, error
  unless ($uid) {
    $self->{msg}= "ERROR: No UID value for updateMemcUIDList";
    carp $self->{msg};
  }

  # get the uid_list
  my $uid_list= $self->getMemcUIDList();

  # if the delete flag was set, then delete the uid from the list
  if ($delete_flag) {
    delete $uid_list->{"uid_$uid"};
  }
  # otherwise, add the uid
  else {
    # this will do nothing if there, add if not 
    $uid_list->{"uid_$uid"}= $uid;
  }

  # store it back in there
  $self->setMemcUIDList($uid_list);
}

# 
# getUsersFromCache 
#   retrieves and returns complete user reference
#   of users from memcached
#
# args
#     $self - object reference
#
#  returns
#     $uref - hashref of user hash refs
#     empty hashref if failure 
#   
#
sub getUsersFromCache {
  my ($self)= @_;
  my $users= {};
  return $users unless defined $self->{memc};

  # get the list of keys for multiple fetching
    my $uid_list= $self->{memc}->get('uid_list');


  #
  # if empty, Cache::Memcached::libmemcached returns an empty string
  # so test whether it's a ref or not as well as being defined
  # if this isn't set, then return, which will in turn
  # force the data to be re-cached
  #
  ref $uid_list or return {};
  keys %$uid_list or return {};

  # the key values of $keys from uid_list privides 
  # the list of keys to get
  $users= $self->{memc}->get_multi(keys %$uid_list);

  # what is needed is for it to be keyed by UID
  for my $uid(keys %$users) {
    $uid =~ /uid_(\d+)/; 
    $users->{$1}= $users->{$uid};
    delete $users->{$uid};
  }

  return $users;
}


=head2 saveUserToCache()

stores a user into cache, insert or update since 'set' doesn't care

=item args

=over

$self - object reference

$uref - user hash ref

=back

=item returns

=over

1 on success

0 failure

=back

=cut

sub saveUserToCache {
  my ($self, $uref)= @_;
  return 0 unless defined $self->{memc};

  #
  # ensure uid_list is updated - this is vitally important
  # this this app working properly!
  #
  $self->updateMemcUIDList($uref->{uid});

  # store the user
  $self->{memc}->set("uid_$uref->{uid}", $uref);

  #
  # other lookup keys - mail and username
  # used in userExists/userExistsInCache
  #
  $self->{memc}->set($uref->{email}, $uref->{uid});
  $self->{memc}->set($uref->{username}, $uref->{uid});
}


=head2 _connectDB()

connects to MySQL, sets the object attribute "dbh", database handle

=item arguments

=over

$self - object reference

=back
 
=item returns

=over

1 - success 

0 - failure

=back

=cut
sub _connectDB {
  my ($self)= @_;

  my $dsn = 'DBI:' . $DB_DRIVER . ':' . $DB_SCHEMA;
  eval {$self->{dbh}= DBI->connect_cached($dsn, $self->{db_user}, $self->{db_password})};
  if ($@) {
    $self->{msg}= "Database ERROR: <font size=\"-3\">$@ $DBI::errstr</font>\n";
    confess $self->{msg};
    return 0;
  }
  unless (defined $self->{dbh}) {
    $self->{msg}= "Database ERROR: <font size=\"-3\">$@ $DBI::errstr</font>\n";
    return 0;
  }
  return 1;
}


=head2 _connectMemcached()

connects to Memcached, sets the object attribute "memc", database handle

=item arguments

=over

$self - object reference

=back

=item returns

=over

1 - success 

0 - failure

=back

=cut

sub _connectMemcached {

  my ($self)= @_;
  $self->{memc}= new Cache::Memcached::libmemcached({
                     servers            =>  $self->{memc_servers},
                     compress_threshold => 10_000 });

  unless ($self->{memc}->set('memcached_alive', 1)) {
    carp "ERROR: Memcached unable to connect!";
    $self->{memc}= undef;
    return 0;
  }
  return 1;
}

 
=head2 getState

returns a full state name given a state abbreviation, using
the state ref returned by getStates()

=item args

=over

$self - object reference

$state_abbr - scalar string state abbreviation value 

=back

=item returns

=over

scalar string state name, success

undef on failure

=back

=cut
sub getState {
  my ($self, $state_abbr)= @_;
  my $states= $self->getStates();
  return $states->{$state_abbr};
}

=head2 getStates()

   method to obtains a hash reference of states - keyed by state 
   abbreviation, value the state name
   
=over args 

$self - object reference

=back

=over returns

hash reference of states keyed by state abbr, value state name

=back

=cut

sub getStates {
  my ($self)= @_;

  my $states = {};
  # a simple true/false to indicate that cities have been cached
  $states= $self->{memc}->get('states_cached') if defined $self->{memc};
  return $states if ref $states;

  # if not retreived via memcached, it has to be ensured it's a hash ref
  # otherwise, strict refs will complain
  $states = {};
  # define the base query
  my $query= 'SELECT state_abbr, state_name FROM states';

  my $sth= $self->{dbh}->prepare($query);

  $sth->execute();
  my $sref= $sth->fetchall_arrayref({});
  $states->{$_->{state_abbr}} = $_->{state_name} for @$sref;

  # indicate that states have been cached
  unless (defined $self->{memc} && $self->{memc}->set('states_cached',$states)) {
    $self->{msg}= "ERROR: setting 'states_cached'\n";
  }
  return $states;

}


=head2 isDenied()

Returns true or false of whether a given IP address is banned or not. 
It uses memcached to fetch the object from memcached for the the 
IP address value, returning the 1 or 0 that it was stored with

=over

=item args

$ip_address - the IP address being checked

=back

=over

=item returns

1 or 0, banned or not

=back

=cut

sub isDenied {
    my ($self, $ip_address)= @_;
    
    my $sth = $self->{dbh}->prepare("select inet_aton(?)");
    $sth->execute($ip_address);
    my $ip_ref= $sth->fetchrow_arrayref();
    $sth->finish();
    return($self->{memc}->get($ip_ref->[0]));
}


=head2 isAuthenticated()

A method used to check if the user and password are in the database,
as supplied when prompted when accessing an <Location> or <Directory> 
that requires it  using md5 digest 

=over

=item Arguments 

$self - object reference

$username - username supplied at prompt

$password - password, in non-md5 form, from prompt

=back

=over

=item Returns

The UID value or undef

=cut

sub isAuthenticated {
    my ($self, $username, $password)= @_;

    # create a hex password using md5_hex, which is how the password is stored
    my $digest = md5_hex($password);

    # create a memcached key by username and hex password
    my $memc_key= $username . '-' . $digest;
    my $uid= $self->{memc}->get($memc_key);
    return($uid) if $uid;

    # user and password must match
    my $query= 'SELECT uid FROM users WHERE username = ? AND password = ?';

    my $resref= $self->{dbh}->selectrow_arrayref($query, 
                                                {},
                                                ($username, $digest));
    # return uid or undef 
    return $resref->[0];
}

=head2 buildUpdate()

Helper method to construct the SET part of an UPDATE statement. This method uses a hash
reference $fields_ref to construct the column names and values specified in the SET
part of an UPDATE statement. Like whereClause, this also allows you to specify literals
so the default placeholder isn't appended for that column as well as deleting the 
value for that column from fields_ref which is inevitably used for the query execution

=item  Args

$self - object reference
$fields_ref - hash reference containing the columns and values for the SET

=item Returns

SET string

=cut
sub buildUpdate {
    my ($self, $fields_ref)= @_; 
    return unless ref $fields_ref && keys %$fields_ref;
    my $set;

    for (keys %$fields_ref) {
        if ($_ =~ /^-/) {
            my $actual_col = $_;
            $actual_col =~ s/^-//;
            $set .= "$actual_col = " . $fields_ref->{$_} . ',';
            delete $fields_ref->{$_};
        }
        else {
            $set .= "$_ = ?,";
        }
    }
    chop($set);

    return " SET  $set "; 
}
=head2 buildInsert()

This is a helper method that constructs the fields and values of an INSERT
SQL statement. It builds the string based off of the keys and values 
of the $fields_ref hash reference

=item Args

$self -- object reference
$fields_ref -- hash reference used to construct the string

=item Returns

String of the form 
"(field1,field2,fieldN...) VALUES (value1, value2, valueN..)" 

=cut

sub buildInsert {
    my ($self, $fields_ref)= @_; 
    my ($values, $fields);

    for (keys %$fields_ref) {
        if ($_ =~ /^-/) {
            $values .= $fields_ref->{$_} . ',';
            delete $fields_ref->{$_};
            $_ =~ s/^-//;
        }
        else {
            $values .= '?,';
        }
        $fields .= $_ . ',';
    }
    chop($fields);
    chop($values);

    return '(' . $fields . ') VALUES (' . $values . ')'; 
}


=head2 whereClause()

Builds up a WHERE clause for a SELECT, DELETE, UPDATE
statement based off of a perl hash reference containing 
the column names with the values of the WHERE condition

=item Arguments

$self - object reference

$fields_ref - hash reference. Fields are columns of the 
where clause, values the values of the where condition
This will build up the where clause with either placeholder
or literals. If literal, it removes that column from 
the hashref so it won't be used in passing placeholders

=item Returns

String, WHERE clause

=cut

sub whereClause {
    my ($self, $fields_ref)= @_; 
    return unless ref $fields_ref && keys %$fields_ref;
    my $where;

    # loop through the reference
    for (keys %$fields_ref) {
        # if the key contains a dash, translate as a literal
        if ($_ =~ /^-/) {
            my $actual_col = $_;
            $actual_col =~ s/^-//;
            $where .= " $actual_col = ". $fields_ref->{$_} . ' AND ';

            # since this won't be used for a place-holder, remove it
            delete $fields_ref->{$_};
        }
        # or append a placeholder
        else {
            $where .= " $_ = ? AND ";
        }
    }

    # remove the trailing AND
    if (defined $where && length($where) ) {
        $where= substr($where, 0, -5);
        $where = "WHERE $where" ;
    }
    return $where;
}

=head2 dbDelete()

Deletes one or more records specified in $where

=item Arguments

$self -- object reference

$table -- table variable

$where -- hash reference used to specify WHERE clause

$options -- any specific options, none for this method yet

=item Returns

Number of rows deleted

=cut

sub dbDelete {
    my ($self, $table, $where, $options) = @_;
    my $rows;

    my $query = "DELETE FROM $table ";

    # build up the where clause
    $query .=  (defined $where && ref($where) eq 'HASH' ) ?
        $self->whereClause($where)  :
        defined $where && length($where) ? $where : '';

    my $sth = $self->runQuery({ 
        'sql'       => $query, 
        'values'    => ref($where) eq 'HASH' ?
            [values %$where] : [] }, \$rows);

    return ($sth->rows());

}

=head2 dbUpdate()

Updates a particular record in the table specified. Uses a hash 
reference to supply the values that will be updated, as well as
another hash reference to supply the WHERE clause

=item Arguments

$self -- object reference

$table -- the table being updated

$fields_ref -- hash reference containing the keys the columns 
and values being updated

$where -- hash reference containing the keys as the columns and values
that will be used in the where clause

=item Returns

number of rows updated

=cut

sub dbUpdate {
    my ($self, $table, $fields_ref, $where, $options) = @_;
    my $rows;

    my $query = "UPDATE $table ";

    # build the SET part of the UPDATE statement
    $query .= $self->buildUpdate($fields_ref) if defined $fields_ref;

    # build the WHERE clause
    $query .= $self->whereClause($where) if defined $where;

    my $sth = $self->runQuery({ 
        'sql'       => $query, 
        'values'    => [values %$fields_ref,values %$where]
        }, \$rows);

    return ($sth->rows());
}



=head2 dbQuery()

Retrieves an array reference of hash references for each record
retrieved from the execution of the query specified

=item Arguments

$self   object reference

$query sql query 

=item Return value

Array reference of hash references for each record


=cut

sub dbQuery {
    my ($self, $query) = @_;

    # run the query, obtain statement handle
    my $sth = $self->runQuery({ 'sql' => $query });

    return $sth->fetchall_arrayref({});
}


=head2 dbGetRef()

Retrieves an array reference of hash references for each record
retrieved from a query specified by list of columns being 
selected, a hashref containing the column names and values 
used to construct a WHERE clause  

=item Arguments

$self   object reference

$table  table name

$fields scalar, comma separated list of columns being selected

$where  hashref, keys column names, values the conditions.  

=item Return value

Array reference of hash references for each record


=cut

sub dbGetRef {
    my ($self, $table, $fields, $where, $options) = @_;
    my $rows;
    $fields||= '*';

    my $query = "SELECT $fields FROM $table ";

    # build up the where clause
    $query .=  (defined $where && ref($where) eq 'HASH' ) ?
        $self->whereClause($where)  :
        defined $where && length($where) ? $where : '';

    # run the query, obtain statement handle
    my $sth = $self->runQuery({ 
        'sql'       => $query, 
        'values'    => ref($where) eq 'HASH' ?
            [values %$where] : [] }, \$rows);

    return $sth->fetchall_arrayref({});
}

=head2 dbInsert()

Inserts a single hash reference, which represents a particular entity into 
the named table. You are responsible for providing the correct name and 
number of columns

=item Arguments

$self -- object reference

$table -- name of the table to be inserted to

$dataref -- hashref, keys column names value values, to be inserted 

$options -- any particular options such as IGNORE 

=item Returns

Id of item inserted from insert id

=cut

sub dbInsert {
    my ($self, $table, $dataref, $options) = @_;

    my $rows;

    # this would cause INSERT IGNORE to be used
    my $ignore= $options->{IGNORE} ? 'IGNORE' : '';

    # build up the SQL INSERT statement
    my $sql = "INSERT $ignore INTO $table " . $self->buildInsert($dataref);

    # run the query
    my $sth = $self->runQuery({ 
            'sql'       => $sql,
            'values'    => [values %$dataref]}, \$rows);

    return(0) unless $sth;

    # return last insert id
    return($sth->{mysql_insertid});
}

sub runQuery {
    my ($self, $params,$rows) = @_;
    my $sth;

    # 3 - remove handle from cache if not being used
    eval { $sth = $self->{dbh}->prepare_cached($params->{sql}, {}, 3); };

    if ($@) {
        warn "ERROR: $@ SQL $params->{sql}";
    }
    eval { $$rows = $sth->execute(@{$params->{values}}); };

    if ($@) {
        warn "ERROR: $@ $params->{sql} bind values\n";        
    }
    return $sth;
}

=head2 getFiles()

Returns an array reference containing hash references for each 
file's meta-data saved in the files table

=item arguments

=over

$self - object reference

$uid - the user ID of the files owner

=back

=item returns

=over

Array reference containing hash references of meta data for every file

=back

=cut

sub getFiles {
    my ($self, $uid) = @_;
    return $self->dbGetRef('files', '*', { uid => $uid});
}

=head2 fileExists()

Returns an either the file id of the entry for a given file's
meta-data or zero, if not exists. The look-up for the data is 
done using the file name (filename) and the user id (uid)

=item arguments

=over

$self - object reference

$fileinfo - hash reference of file meta data 

=back

=item returns

=over

file id - success
0       - failure

=back

=cut

sub fileExists {
    my ($self, $fileinfo)= @_;

    my $file_exists= $self->dbGetRef('files', 'id', { 
        uid         => $fileinfo->{uid},
        filename    => $fileinfo->{filename}});

    # file id, primary key of file
    return(defined $file_exists->[0]{id} ? $file_exists->[0]{id} : 0 );
}

=head2 saveFileInfo()

Either updates or inserts a given file's meta data, depending on 
whether the meta-data entry already existed or not

=item arguments

=over

$self - object reference

$fileinfo - hash reference of file meta data 

=back

=item returns

=over

$rows
> 0     - success
0       - failure

=back

=cut

sub saveFileInfo {
    my ($self, $fileinfo)= @_;
    my $file_id= $self->fileExists($fileinfo);
    my $rows;
    
    # if file exists, update
    if ($file_id) {
        $rows = $self->dbUpdate('files', $fileinfo, { id => $file_id });
    }
    # otherwise update
    else {
        $rows = $self->dbInsert('files', $fileinfo); 
    }
    return $rows;
}

sub randomString {
  my @chars = grep !/[0O1Iil]/, 0..9, 'A'..'Z', 'a'..'z';
  return join '', map { $chars[rand @chars] } 0 .. 7;
}

=head2 isAdmin()

This method returns 1 or 0, representing if the user is 
an admin user or not

=over

=item args

$self - instantiated object ref to itself 

$username - the username to check

=back

=over

=item returns

1 - is admin

0 - is not admin

=back

=back

=cut

sub isAdmin {
    my ($self, $username)= @_;
    my $user= $self->getUser({username => $username});
    return defined $user && $user->{admin} == 1 ? 1 : 0;
}


=head2 cacheBannedIPs()

cache the banned_ips table 

=over

=item args

$self - object reference

=back

=over

=item returns 

no value

=back

=cut

sub cacheBannedIPs {
    my ($self)= @_;

    # fetch all data from banned_ips
    my $query= 'SELECT ip_address, banned FROM banned_ips';

    my $bref= $self->{dbh}->selectall_arrayref($query);

    # cache by IP address
    $self->{memc}->set($_->[0], $_->[1]) for @$bref;

    return;
}


=head2 weblog()

Logs apache log entries to MySQL

=over

=item Args

$self           - object reference

$logref         - a hash reference containing the following:

=over
uri             - the URI of the request

hostname        - the hostname of the client

ip_address      - the ip address of the client

method          - the request method

user            - the user id of the client, if set

filename        - the mapped filename of the URI

useragent       - the user-agent value

last_modifed    - last modification date of the resource

status          - status of response

=back
 
=back

=over

=item Returns 

1 for inserted, 0 for not 

=back

=cut   

sub weblog {
    my ($self,$logref)= @_;

    #
    # $logfields is used from the mapping of field value in query 
    # generation and bind values to placeholder 
    #
    my @logfields= qw(
                        uri
                        hostname
                        method
                        userid
                        useragent
                        filename
                        status
                        ip_address
                        last_modified
                        );

    my $insert= 'INSERT INTO weblog (';

    # build up the fields list 
    $insert .= join(',', @logfields);
    $insert .= ') VALUES (';

    #
    # build up the placeholders 
    # -2 because inet_aton and from_unixtime need to be used
    #
    $insert .= '?,' x (scalar @logfields - 2);
    $insert .= ' inet_aton(?), ';
    $insert .= ' from_unixtime(?) )';

    my $sth = $self->{dbh}->prepare($insert);

    #
    # use a hash slice to pass the bind variables- the values of
    # $logfields ensures order of placeholders and their values
    #
    my $inserted= $sth->execute(@{$logref}{@logfields});

    return $sth->rows();

}

sub makeSessionId {
    my ($uid)= @_;
    my $value= randomString();
    $value.= localtime;
    $value= md5_hex($value);
    return $value;
}

sub encodeCookieValue {
    my ($id, $value) = @_;
    $value = $id . '::' . $value;
    $value =~ s/(.)/sprintf("%%%02x", ord($1))/ge;
    return $value;
}

sub digestCookieValue {
    my ($value) = @_;
    $value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack('C', hex($1))/ge;
    my ($id, $sessionid) = split(/::/, $value, 2);
    return({
        id          => $id,
        sessionid   => $sessionid});
}

sub insertSession {
    my ($self, $sessionid, $uid)= @_;
    
    my $sessionref = { 
        uid         => $uid, 
        sessionid   => $sessionid, 
        user        => $self->getUser({ uid => $uid})
        };

    $self->dbInsert('sessions', {
            sessionid       => $sessionid,
            uid             => $uid,
            stored_session  => nfreeze($sessionref),
            }
        );
    $self->{memc}->set($sessionid, $sessionref); 
    return $sessionref;
}

sub storeSession {
    my ($self, $sessionid, $newsession)= @_;

    my $existingsession= $self->getSessionData($sessionid);
    @{$existingsession}{keys %$newsession}= values %$newsession;

    $self->dbUpdate('sessions', 
            { stored_session => nfreeze($existingsession)},
            { sessionid     => $sessionid});

    $self->{memc}->set($sessionid, $existingsession);

    return $existingsession;
};


sub getSessionData {
    my ($self, $sessionid)= @_;
    my $sessionref;

    # check in memcached
    $sessionref= $self->{memc}->get($sessionid);

    # return if in memcached and has a session id
    return $sessionref if $sessionref->{sessionid};

    # otherwise, obtain from database
    my $sref= $self->dbGetRef('sessions', 'sessionid, uid, stored_session',
                                        { sessionid   => $sessionid, });

    # if nothing, return empty hash ref
    return {} unless defined $sref && scalar @$sref;

    # de-serialize
    $sessionref= thaw($sref->[0]{stored_session});

    # write to memcached on the way out
    $self->{memc}->set($sessionid, $sessionref);

    # finally, return the session ref
    return $sessionref;
}

sub unsetCookie {
    my ($r)= @_;
    setCookie($r, '');
}

sub setCookie {
    my ($r, $value)= @_;

    # if no value, then this means a delete cookie
    my $expires= length($value) ? '+3M' : '-3M';

    my $cookie = Apache2::Cookie->new($r,
        -name    =>  $COOKIENAME,
        -value   =>  $value, 
        -expires =>  $expires,
        -domain  =>  $COOKIEDOMAIN, 
        -path    =>  $COOKIEPATH,
        );
    $cookie->bake($r);
}

sub getSession {
    my ($self, $msg)= @_;

    my $j = new Apache2::Cookie::Jar($self->{r});

    my $cookie= $j->cookies($COOKIENAME);

    if ($cookie && $cookie->value()) {
        my $cookieref= digestCookieValue($cookie->value()); 
        if ($self->{form}{'logout'}) {
            unsetCookie($self->{r});
            $self->deleteSession($cookieref->{sessionid});
            $$msg= 'User is logged out';
            return {};
        }
        else {
            my $sessionref= $self->getSessionData($cookieref->{sessionid});
            unless ($sessionref->{sessionid}) {
                unsetCookie($self->{r});
                return {};
            }
            return $sessionref;
        }
    }
    elsif ($self->{form}{login}) {
        if (my $uid= $self->isAuthenticated(
                $self->{form}{'username'}, 
                $self->{form}{'password'})) {

            my $sessionid= $self->makeSessionId($uid);

            my $sessionref= $self->insertSession($sessionid, $uid);

            setCookie($self->{r}, encodeCookieValue($uid, $sessionid));

            $$msg= "Session was created";

            return $sessionref;
        }
        else {
            $$msg= "User " . $self->{form}{username} . " not authenticated. Try again";
            return {};
        }
    }
}

sub deleteSession {
    my ($self, $sessionid)= @_;
    $self->dbDelete('sessions', {sessionid => $sessionid});
    $self->{memc}->delete($sessionid);
}


#
# cacheUsersByMD5
#   cache all the users by username-md5passowrd
#   used for authentication lookups
#
# args
#   object hash reference 
#
# returns
#   nothing
#   
sub cacheUsersByMD5 {
    my ($self)= @_;
    my $query= "SELECT concat(username, '-', password), uid FROM users";

    my $uref = $self->{dbh}->selectall_arrayref($query);

    map {$self->{memc}->set($_->[0], $_->[1])} @$uref;

    return;
}

=head2 getForm()

Returns a hash reference containing the submitted form values

=item Args

$self -- object reference

=item Returns

hash reference with form values 

=cut
sub getForm {
    my $self= $_[0];
    return $self->{form};
}

sub buildRange {

    my ( $results_per_page, $total ) = @_;

    my @ranges;
    my $start = 1;
    while ( $start + $results_per_page < $total ) {

      push @ranges, [ $start, $start + $results_per_page - 1 ];
      $start += $results_per_page;
    }

    # this handles the last range
    $start > $total or push @ranges, [ $start, $total ];

    return \@ranges;
}

sub buildRangeOld {
    my ($results_per_page, $total) = @_;
    return [[1, $total]] if ($total <= $results_per_page);
    
    # used for collecting the values
    my $start;
    my $end;
    my $ranges;

    # get the total number of pages for the result set 
    my $pages= $total/$results_per_page;

    # found out the remainder - the last set 
    my $remainder= $total % $results_per_page;

    # loop through the number of pages 
    for my $i(1 .. $pages) {
        # end 
        $end = $results_per_page * $i;
        # start
        $start = $end - $results_per_page + 1;
        # stash the set
        push(@$ranges, [$start, $end]);
    }

    # now tack on the remaining set
    $start = $end + 1;
    $end += $remainder;
    push(@$ranges, [$start, $end]);
    return $ranges;
}

sub sendEmail {
  my ($addr, $from, $subject, $content, $emailhost) = @_;
  $from= 'accounts@patg.net';
  $emailhost ||= 'patg.net'; 

  unless (Email::Valid->rfc822($addr)) {
    warn "Can't send mail '$subject' to $addr: Invalid address\n";
    return 0;
  }

  my %data = (
      From            => $from,
      Smtp            => $emailhost,
      Subject         => $subject,
      Message         => $content,
      To              => $addr,
      'Content-type'  => 'text/plain; charset="us-ascii"',
      'Content-transfer-encoding' => '8bit'
      );

  if (Mail::Sendmail::sendmail(%data)) {
    return 1;
  }
  else {
    warn "Can't send mail '$subject' to $addr: $Mail::Sendmail::error\n";
    return 0;
  }

}
sub cacheTemplate {
    my ($self, $template_path, $template_name)= @_;

    my $template_content= '';
    my $template_key = "templates:$template_name";

    $template_content = $self->{memc}->get($template_key);
    if ($template_content) {
        warn "template is cached.\n";
        return $template_content; 
    }
    
    warn "template not cached, caching.";
    $template_content = read_file("$template_path/$template_name") ;

    $self->{memc}->set($template_key,$template_content);
    return $template_content;
}

=head2 encodeUserData()

Encodes any HTML entities which might be contained in each 
user attribute to prevent HTML injection in any forms or pages
that print out this data

=item arguments 

=over

$self - object reference

$user - reference to user hash reference

=back

=item returns

=over

does not return a value since it modifies $user passed as a reference

=back

=cut

sub encodeUserData {
    my ($self, $user)= @_;

    # encode any HTML entities if defined
    for my $attr (@$USER_COLS) {
        if (defined $$user->{$attr}) {
            $$user->{$attr}= encode_entities($$user->{$attr}) ;
        }
    }
    return;
}

1;
