package Narada::SearchHandler;

use strict;

use Apache2::Const -compile => qw(OK REDIRECT SERVER_ERROR);
use Apache2::Request;
use Template; 
use Sphinx::Search;
use Getopt::Long;
use Data::Dumper;

use Narada;
use CGI qw(:standard);


sub handler {
    my ($r)= @_;
    my $msg;
    my $url= $r->uri;
    my $template_path;
    my $sinfo;

    # instantiate a Narada object
    my $narada= new Narada({ r => $r});

    # get the submitted form values
    my $form= $narada->getForm();

    # defaults 
    $form->{results_per_page} = $form->{results_per_page} ?
                            $form->{results_per_page} : 10;

    $form->{start} ||= 1;
    $form->{end} ||= 10;

    my $res_pp_select = popup_menu(
                          -name     => 'results_per_page',
                          -values   => [10, 25, 50, 100],
                          -onchange => 'this.form.submit();',
                          -default  => $form->{results_per_page} ? 
                                $form->{results_per_page} : 10);



    my $sessionref= $narada->getSession(\$msg);

    # if no session, redirect to login page
    unless ($sessionref && $sessionref->{uid}) {
        $r->headers_out->add('Location' => "/login?returnto=$url");
        return Apache2::Const::REDIRECT;
    }

    # obtain the template path
    $template_path= $r->dir_config('TEMPLATE_PATH');

    # allow select box to resubmit when results per set changes
    $form->{query} ||= $form->{hquery};

    # if the user has submitted a search, get the results
    if ($r->method() eq 'POST' || $r->method() eq 'GET' && $form->{query}) {
        $sinfo = search($r, $narada, $form);
    }

    if (defined $sinfo && defined $sinfo->{sresults} && scalar @{$sinfo->{sresults}}) {
        $sinfo->{ranges}= buildRange(
                            $form->{results_per_page},
                            $sinfo->{total_found});
    }

    # instantiate a new Template object
    my $template = Template->new({
        INCLUDE_PATH => $template_path,  # where to look for templates
        POST_CHOMP   => 1,               # removal of whitespace
    });

    # set up template parameters
    my $tparams = {
        msg             => "Session " . $sessionref->{sessionid} . " logged in",
        url             => $r->uri,
        sinfo           => $sinfo,
        form            => $form,
        res_pp_select   => $res_pp_select,
    };

    # set up the template name
    $url =~ s/^\///;
    my $tname= $url . '.tt2';
    my $template_cached= $narada->cacheTemplate($template_path,$tname); 

    if ($template_cached) {
        $tname = \$template_cached;
    }

    $r->content_type('text/html');


    # this will process the template
    $template->process($tname, $tparams) ||
        do {
            # if an error, return SERVER_ERROR
            $r->server->warn($template->error());
            return Apache2::Const::SERVER_ERROR;
        };

    return Apache2::Const::OK;
}

sub search {
    my ($r, $narada, $form)= @_;
    my $sinfo;
    my $cref;
    my $docs= [];

    my $search_index = $r->dir_config('SEARCH_INDEX');
    my $excerpts_index = $r->dir_config('EXCERPT_INDEX');
    my $search_host = $r->dir_config('SEARCH_HOST');
    my $search_port = $r->dir_config('SEARCH_PORT');

    my $sphinx = Sphinx::Search->new();

    $sphinx->SetServer($search_host, $search_port);
    $sphinx->SetRankingMode(SPH_RANK_PROXIMITY_BM25);
    $sphinx->SetLimits($form->{start}, $form->{end});


    # other options for extending functinality
    $sphinx->SetMatchMode(SPH_MATCH_EXTENDED2);
    # this provides ordering, but ordering is lost from the db.
    #$sphinx->SetSortMode(SPH_SORT_EXTENDED,'ulast_updated') ;

    my $hits= $sphinx->Query($form->{query}, $search_index);
    my $warning = $sphinx->GetLastWarning;

    $hits->{total_found} or return {};

    $sinfo->{$_} = $hits->{$_} for qw (total_found time);

    $sinfo->{sresults} = get_results_from_db($narada, $hits); 

    my $warning = $sphinx->GetLastWarning;
    $r->server->warn($warning) if $warning;

    my $term= $form->{query};
    $term =~ s/[|()"]/ /g;
    $term=~s/\~1//g;
    $term =~ s/\s{2,}/ /g;

    if (scalar @{$sinfo->{sresults}}) {
        push(@$cref, $_->{content}) for @{$sinfo->{sresults}};
        my $excerpts = $sphinx->BuildExcerpts(
            $cref,
            $excerpts_index,
            $term,
            {
            before_match    => '<em>',
            after_match    => '</em>',
            around          => 3,
            single_passage  => 0,
            limit           => 180});

        # add in the results
        for my $i (0 .. $#{$sinfo->{sresults}}) { 
            $sinfo->{sresults}[$i]{excerpt}= $excerpts->[$i];
            delete $sinfo->{sresults}[$i]{content};
        }
    }
    return $sinfo;
}

sub get_results_from_db {
    my ($narada, $hits)= @_;

    # build up WHERE .. IN
    my $where = 'WHERE id IN (';

    $where .= join(',', map { $_->{doc} } @{$hits->{matches}});
    $where.= ')';
    $where.= ' ORDER BY last_modified DESC';

    my $time_format= q(date_format(last_updated, '%W, %m/%d/%y %H:%i:%s') as last_updated);
    my $sresults= $narada->dbGetRef('urls join urls_blob using (id)',
                "urls.id as id, title, $time_format, url, content", $where);

    return $sresults;
}

1;
