<?php
# Narada Search Engine in PHP
# Copyright (C) 2009 Eric Day, Patrick Galbraith
# All rights reserved.
#
# Use and distribution licensed under the BSD license.  See
# the COPYING file in this directory for full text.

# This is a helper/base class for all Narada components to use. It encapsulates
# common configuration and initialization of things like Gearman clients and
# workers.

require('NaradaConf.php');

#if ($GLOBALS['NaradaDebug'])
#  error_reporting(E_ALL | E_NOTICE | E_STRICT);

class Narada
{
  private $gearmanServers;
  private $gearmanClient;
  private $drizzle;
  private $drizzleCon;
  private $sphinx;
  private $sphinxIndex;
  private $sphinxHost;
  private $sphinxPort;
  private $sphinxMatchMode;
  protected $functionName;
  protected $fetchFunctionName;
  protected $searchFunctionName;
  protected $insertFunctionName;
  protected $indexFunctionName;

  public function __construct()
  {
    $servers= explode(',', $GLOBALS['NaradaGearmanServers']);
    foreach ($servers as $server)
      $this->gearmanServers[]= explode(':', $server);

    $this->gearmanClient= new GearmanClient();

    foreach ($this->gearmanServers as $server)
      $this->gearmanClient->addServer($server[0], $server[1]);

    $this->fetchFunctionName= $GLOBALS['NaradaFetchFunctionName'];
    $this->searchFunctionName= $GLOBALS['NaradaSearchFunctionName'];
    $this->insertFunctionName= $GLOBALS['NaradaInsertFunctionName'];
    $this->indexFunctionName= $GLOBALS['NaradaIndexFunctionName'];

    $this->sphinx= NULL;
    $this->sphinxIndex= $GLOBALS['NaradaSearchIndex'];
    $this->sphinxHost= $GLOBALS['NaradaSearchHost'];
    $this->sphinxPort= $GLOBALS['NaradaSearchPort'];
    $this->sphinxMatchMode= $GLOBALS['NaradaSearchMode'];

    $this->drizzle= new Drizzle();
    $this->drizzleCon= $this->drizzle->addTcp($GLOBALS['NaradaDrizzleHost'],
                                              $GLOBALS['NaradaDrizzlePort'],
                                              $GLOBALS['NaradaDrizzleUser'],
                                              $GLOBALS['NaradaDrizzlePass'],
                                              $GLOBALS['NaradaDrizzleDb'],
                                              $GLOBALS['NaradaDrizzleOptions']);
  }

  public function debug($msg)
  {
    if ($GLOBALS['NaradaDebug'])
      print "$msg\n";
  }

  public function insert($url)
  {
    $this->gearmanClient->doBackground($this->insertFunctionName, $url);
  }

  public function search($query)
  {
    return $this->gearmanClient->do($this->searchFunctionName, $query);
  }

  protected function fetch($url)
  {
    $this->gearmanClient->doBackground($this->fetchFunctionName, $url);
  }

  public function index()
  {
    $this->gearmanClient->doBackground($this->indexFunctionName, NULL);
  }

  protected function store($url)
  {
    $result= @$this->drizzleCon->query("INSERT INTO url SET url_md5='" .
                                       md5($url) . "', url='" .
                                       $this->drizzle->escapeString($url) . "'");
    if ($result)
      return $result->insertId();

    return 0;
  }

  protected function storeContent($id, $title, $content)
  {
    $query= "INSERT INTO url_content SET id=$id, title='" .
            $this->drizzle->escapeString($title) .
            "', content='" .
            $this->drizzle->escapeString($content) .
            "' ON DUPLICATE KEY UPDATE content='" .
            $this->drizzle->escapeString($content) . "'";

    $this->drizzleCon->query($query);
  }

  protected function get($id)
  {
    $result= $this->drizzleCon->query("SELECT url FROM url WHERE id=$id");
    if (!$result || $result->errorCode() || $result->buffer())
      return NULL;

    $row= $result->rowNext();
    return $row[0];
  }

  protected function getContent($search_results)
  {
    if (!is_array($search_results["matches"]))
      return NULL;

    $query= 'SELECT id, url, title, content ' .
            'FROM url JOIN url_content USING (id) ' .
            'WHERE id IN (' .
            implode(",", array_keys($search_results["matches"])) . ')';

    $result= $this->drizzleCon->query($query);
    if (!$result || $result->errorCode() || $result->buffer())
      return NULL;

    while($row= $result->rowNext())
    {
      $results[$row[0]]["url"]= $row[1];
      $results[$row[0]]["url_title"]= $row[2];
      $results[$row[0]]["url_content"]= $row[3];
    }

    return $results;
  }

  protected function searchIndex($query)
  {
    if ($this->sphinx == NULL)
    {
      $this->sphinx= new SphinxClient();
      $this->sphinx->SetServer($this->sphinxHost, $this->sphinxPort);
      $this->sphinx->SetMatchMode($this->sphinxMatchMode);
    }

    return $this->sphinx->Query($query, $this->sphinxIndex);
  }

  protected function verify($url)
  {
    if (strstr($url, "://"))
    {
      if (substr($url, 0, 7) != "http://" && substr($url, 0, 8) != "https://")
        return NULL;
    }
    else
      $url= "http://$url";

    if (substr_count($url, "/") == 2)
      $url= "$url/";

    return $url;
  }

  # Derived worker classes should set functionName, define a callback member
  # function, and then call this.
  public function runWorker()
  {
    $worker= new GearmanWorker();

    foreach ($this->gearmanServers as $server)
      $worker->addServer($server[0], $server[1]);

    $worker->addFunction($this->functionName, "NaradaWorkerCallback", $this);

    while ($worker->work());
  }
};

# Callback function for Gearman worker interface.
function NaradaWorkerCallback($job, &$narada)
{
  return $narada->callback($job);
}

?>
